#NO_APP
	.text
	.align 1
	.globl _bn_sub_words
_bn_sub_words:
	;; register usage:
	;;   r10 - BN_ULONG *r, the result array
	;;   r11 - BN_ULONG *a, source array a
	;;   r12 - BN_ULONG *b, source array b
	;;   r13 - int n , the length of the arrays
	;;   r0 to r7, register bank
	;;   r8 carry save
	;;   r9 tmp num counter

	;; put r0 to r8 on stackframe (9*4==36 octets)
	subq 36,sp
	movem r8,[sp]	
	clearf c
	move dccr,r8
	subq 8,r13		; are there 8 elems or more?
	blt LSUB_UNROLL_START
	nop
	
LSUB_LOOP_START:
	movem [r11+],r7
	;; restore dccr to continue subtraction
	move r8,dccr
	ax
	sub.d [r12+],r7
	ax
	sub.d [r12+],r6
	ax
	sub.d [r12+],r5
	ax
	sub.d [r12+],r4
	ax
	sub.d [r12+],r3
	ax
	sub.d [r12+],r2
	ax
	sub.d [r12+],r1
	ax
	sub.d [r12+],r0
	;; save dccr to r8 (to get the carry flag)
	move dccr,r8
	movem r7,[r10+]	; move the result to r array

	subq 8,r13		; are there 8 elems or more left?
	bge LSUB_LOOP_START
	nop

LSUB_UNROLL_START:
	addq 8,r13
	ble LSUB_END
	nop
LSUB_UNROLL:
	move.d [r11+],r0
	move.d [r12+],r1
	move r8,dccr
	ax
	sub.d r1,r0
	move dccr,r8
	move.d r0,[r10+]
	subq 1,r13
	bgt LSUB_UNROLL
	nop
LSUB_END: 
	;; done, set the return value and restore registers
	move r8,dccr
	ax
	slo r10
	movem [sp+],r8
	Ret
	nop

	.align 1
	.globl _bn_add_words
_bn_add_words:
	;; register usage:
	;;   r10 - BN_ULONG *r, the result array
	;;   r11 - BN_ULONG *a, source array a
	;;   r12 - BN_ULONG *b, source array b
	;;   r13 - int n , the length of the arrays
	;;   r0 to r7, register bank
	;;   r8 carry save
	;;   r9 tmp num counter

	;; put r0 to r8 on stackframe (9*4==36 octets)
	subq 36,sp
	movem r8,[sp]
	clearf c
	move dccr,r8
	subq 8,r13		; are there 8 elems or more?
	blt LADD_UNROLL_START
	nop
	
LADD_LOOP_START:
	movem [r11+],r7
	;; restore dccr to continue addition
	move r8,dccr
	ax
	add.d [r12+],r7
	ax
	add.d [r12+],r6
	ax
	add.d [r12+],r5
	ax
	add.d [r12+],r4
	ax
	add.d [r12+],r3
	ax
	add.d [r12+],r2
	ax
	add.d [r12+],r1
	ax
	add.d [r12+],r0
	;; save the dccr to r8 (to get the carry flag)
	move dccr,r8
	movem r7,[r10+]		; move the result to r array

	subq 8,r13		; are there 8 elems or more left?
	bge LADD_LOOP_START
	nop

LADD_UNROLL_START:
	addq 8,r13
	ble LADD_END
	nop	
LADD_UNROLL:
	move.d [r11+],r0
	move.d [r12+],r1
	move r8,dccr
	ax
	add.d r1,r0
	move dccr,r8
	move.d r0,[r10+]
	subq 1,r13
	bgt LADD_UNROLL
	nop
LADD_END: 
	;; done, set the return value and restore registers
	clear.d r10
	move r8,dccr
	ax
	addq 0,r10
	movem [sp+],r8
	Ret
	nop
	.align 1
	.globl _bn_mul_add_words
_bn_mul_add_words:
	;; calculate <r> = <a>*w + <r>
	;; register usage:
	;;   r10 - BN_ULONG *r, the result and additive source array
	;;   r11 - BN_ULONG *a, the multiplicative source array
	;;   r12 - int num, the length of the r & a arrays
	;;   r13 - BN_ULONG w, the multiplicative scalar
	;;   r4 work copy of r10 (<r>)
	;;   r5 work copy of r11 (<a>)
	;;   r6 copy of w
	;;   r7 copy of num
	;; 
	;;   r8 carry dword (for loop wrap)
	;; r9 tmp reg for add, also used by multiplication routine
	;; r10 thru r13 used for subroutine calling
	;; 

	;; put r0 to r8 on stackframe (9*4==36 octets)
	push srp
 	subq 36,sp
	movem r8,[sp]
	;; move indata to r4-r7 since we need r10-r13 for subroutine calls
	move.d r10, r4		; <r>
	move.d r11, r5		; <a>
	move.d r12, r7		; num
	move.d r13, r6		; w

	subq 4,r7
	blt LMUL_ADD_LOOP_ENTRY2
	clear.d r8

LMUL_ADD_LOOP1:	
	movem [r4],r3		; read four elems from <r> to register bank

	move.d [r5+],r10
	move.d r6,r11
  	jsr Lcris_bn_asm_mul_32_to_64
	add.d r10,r3
	ax
	addq 0,r11
	add.d r8,r3		; add in carry dword
	ax
	addq 0,r11		; add in carry bit in hi dword
	move.d r11,r8

	move.d [r5+],r10
	move.d r6,r11
	jsr Lcris_bn_asm_mul_32_to_64
	add.d r10,r2
	ax
	addq 0,r11
	add.d r8,r2		; add in carry dword
	ax
	addq 0,r11		; add in carry bit in hi dword
	move.d r11,r8

	move.d [r5+],r10
	move.d r6,r11
	jsr Lcris_bn_asm_mul_32_to_64
	add.d r10,r1
	ax
	addq 0,r11
	add.d r8,r1		; add in carry dword
	ax
	addq 0,r11		; add in carry bit in hi dword
	move.d r11,r8
	
	move.d [r5+],r10
	move.d r6,r11
	jsr Lcris_bn_asm_mul_32_to_64
	add.d r10,r0
	ax
	addq 0,r11
	add.d r8,r0		; add in carry dword
	ax
	addq 0,r11		; add in carry bit in hi dword
	move.d r11,r8

	movem r3,[r4+]		; write back to <r>

	subq 4,r7
	bge LMUL_ADD_LOOP1
	nop
	
LMUL_ADD_LOOP_ENTRY2:
	addq 4,r7
	beq LMUL_ADD_END
	nop
LMUL_ADD_LOOP2:		
	;; setup arguments for multiplication call
	;;    arg 1 (32 bit unsigned) in r10
	;;    arg 2 (32 bit unsigned) in r11
	;;    result (64 bit unsigned) in r10,r11 (low,high)
	move.d [r5+],r10
	move.d r6, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	move.d [r4],r9
	add.d r10,r9
	ax
	addq 0,r11
	add.d r8,r9		; add in carry dword
	ax
	addq 0,r11		; add in carry bit in hi dword

	move.d r9,[r4+]
	
	subq 1,r7
	bgt LMUL_ADD_LOOP2
	move.d r11,r8		; move carry dword to r8

LMUL_ADD_END:
	move.d r8, r10
	;;  do we have a carrybit from dccr to worry about here?
	movem [sp+],r8
	jump [sp+]
	
	.align 1
	.globl _bn_mul_words
_bn_mul_words:
	;; calculate <r> = <a>*w
	;; register usage:
	;;   r10 - BN_ULONG *r, the result array
	;;   r11 - BN_ULONG *a, the source array
	;;   r12 - int num, the length of the r & a arrays
	;;   r13 - BN_ULONG w, the multiplicative scalar
	;;   r4 work copy of r10 (<r>)
	;;   r5 work copy of r11 (<a>)
	;;   r6 copy of w
	;;   r7 copy of num
	;; 
	;;   r8 carry dword (for loop wrap)
	;; r9 tmp reg for add, also used by multiplication subroutine
	;; r10 thru r13 used for subroutine calling
	;; 

	;; put r0 to r8 on stackframe (9*4==36 octets)
	push srp
	subq 36,sp
	movem r8,[sp]
	;; move indata to r4-r7 since we need r10-r13 for subroutine calls
	move.d r10, r4		; <r>
	move.d r11, r5		; <a>
	move.d r12, r7		; num
	move.d r13, r6		; w

	subq 4,r7
	blt LMUL_LOOP_2_ENTRY
	clear.d r8
	
LMUL_LOOP_1:
	movem [r5+],r3

	move.d r3,r10
	move.d r6,r11
	jsr Lcris_bn_asm_mul_32_to_64

	add.d r8,r10
	ax
	addq 0,r11
	move.d r10,r3
	move.d r11,r8
	
	move.d r2,r10
	move.d r6,r11
	jsr Lcris_bn_asm_mul_32_to_64

	add.d r8,r10
	ax
	addq 0,r11
	move.d r10,r2
	move.d r11,r8

	move.d r1,r10
	move.d r6,r11
	jsr Lcris_bn_asm_mul_32_to_64
	
	add.d r8,r10
	ax
	addq 0,r11
	move.d r10,r1
	move.d r11,r8

	move.d r0,r10
	move.d r6,r11
	jsr Lcris_bn_asm_mul_32_to_64

	add.d r8,r10
	ax
	addq 0,r11
	move.d r10,r0

	movem r3,[r4+]

	subq 4,r7
	bge LMUL_LOOP_1
	move.d r11,r8
LMUL_LOOP_2_ENTRY:
	addq 4,r7
	beq LMUL_END
	nop
LMUL_LOOP_2:
	;; setup arguments for multiplication call
	;;    arg 1 (32 bit unsigned) in r10
	;;    arg 2 (32 bit unsigned) in r11
	;;    result (64 bit unsigned) in r10,r11 (low,high)
	move.d [r5+],r10
	move.d r6, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	add.d r8,r10		; add in carry dword
	ax
	addq 0,r11		; add in carry bit in hi dword
	move.d r10,[r4+]
	
	subq 1,r7
	bgt LMUL_LOOP_2
	move.d r11,r8		; move carry dword to r8

LMUL_END:
	move.d r8, r10
	movem [sp+],r8
	jump [sp+]

	.align 1
	.globl _bn_sqr_words
_bn_sqr_words:
	;; calculate <r> = <a>^2, r is 2*num long, a is num long
	;; register usage:
	;;   r10 - BN_ULONG *r, the result and additive source array
	;;   r11 - BN_ULONG *a, the multiplicative source array
	;;   r12 - int num, the legnth of the r & a arrays
	;;   r6 work copy of r10 (<r>)
	;;   r7 work copy of r11 (<a>)
	;;   r8 copy of num
	;; 
	;;   r0 to r5, register bank
	;; 
	;; r9 tmp reg for loop counter, also used by multiplication subroutine
	;; r10 thru r13 used for subroutine calling
	;; 

	;; put r0 to r8 on stackframe (9*4==36 octets)
	push srp
	subq 36,sp
	movem r8,[sp]
	
	;; move indata to r6-r8 since we need r10-r13 for subroutine calls
	move.d r10, r6
	move.d r12, r8

	test.d r8
	ble LSQR_END
	
	move.d r8,r9
	subq 6,r9
	blt LSQR_END_LOOP
	move.d r11, r7

LSQR_LOOP:
	move.d r9,r8
	movem [r7+],r5

	move.d r5,r10
	move.d r10, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	move.d r10,[r6+]
	move.d r11,[r6+]
	
	move.d r4,r10
	move.d r10, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	move.d r10,[r6+]
	move.d r11,[r6+]

	move.d r3,r10
	move.d r10, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	move.d r10,[r6+]
	move.d r11,[r6+]

	move.d r2,r10
	move.d r10, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	move.d r10,[r6+]
	move.d r11,[r6+]

	move.d r1,r10
	move.d r10, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	move.d r10,[r6+]
	move.d r11,[r6+]

	move.d r0,r10
	move.d r10, r11
	Jsr Lcris_bn_asm_mul_32_to_64
	move.d r10,[r6+]
	move.d r11,[r6+]

	move.d r8,r9
	subq 6,r9
	bge LSQR_LOOP

	test.d r8
	ble LSQR_END
	nop

LSQR_END_LOOP:
	move.d [r7+],r10
	move.d r10, r11
	Jsr Lcris_bn_asm_mul_32_to_64

	move.d r10,[r6+]
	move.d r11,[r6+]
	subq 1,r8
	bgt LSQR_END_LOOP
	nop

LSQR_END:
	movem [sp+],r8
	jump [sp+]
	.align 1
	.globl _bn_mul_comba4
_bn_mul_comba4:
	;; calculate <r> = <a>*<b>, r is 8 long, a,b are 4 long
	;; register usage:
	;;   r10 - BN_ULONG *r, the result and additive source array
	;;   r11 - BN_ULONG *a, the multiplicative source array
	;;   r12 - BN_ULONG *b, the multiplicative source array
	;;   r0 work copy of r10 (<r>)
	;;   r1 work copy of r11 (<a>)
	;;   r2 work copy of r12 (<b>)
	;;   r3 thru r5 carry registers
	;; 
	;;   r6 thru r8 not used
	;; 
	;; r10 thru r13 used for subroutine calling
	;; 

	;; put r0 to r5 on stackframe (6*4==24 octets)
	push srp
	subq 24,sp
	movem r5,[sp]
	;; move indata to r0-r3 since we need r10-r13 for subroutine calls
	move.d r10, r0
	move.d r11, r1
	move.d r12, r2

	.macro mul_add_c p1,p2,c0,c1,c2
	;; this is basically the mul_add_c macro in asm form, with
	;;    c0 in r3
	;;    c1 in r4
	;;    c2 in r5
	;;    t1 in r10 (multiplication low dword out)
	;;    t2 in r11 (multiplication high dword out)
	;;   t is not needed since we get multiplication output in two regs
	move.d \p1,r10
	move.d \p2,r11
	jsr Lcris_bn_asm_mul_32_to_64

	add.d r10,\c0		  ;
	ax
	addq 0,r11		  ; add in carry, cannot overflow due to
	;;                          size of mul result
	add.d r11,\c1		  ;
	ax
	addq 0,\c2
	.endm
	
	clear.d r3
	clear.d r4
	clear.d r5
	mul_add_c [r1],[r2],r3,r4,r5
	move.d r3, [r0+]
	clear.d r3

	mul_add_c [r1],[r2+4],r4,r5,r3
	mul_add_c [r1+4],[r2],r4,r5,r3
	move.d r4, [r0+]
	clear.d r4

	mul_add_c [r1+8],[r2],r5,r3,r4
	mul_add_c [r1+4],[r2+4],r5,r3,r4
	mul_add_c [r1],[r2+8],r5,r3,r4
	move.d r5,[r0+]
	clear.d r5

	mul_add_c [r1+12],[r2],r3,r4,r5
	mul_add_c [r1+8],[r2+4],r3,r4,r5
	mul_add_c [r1+4],[r2+8],r3,r4,r5
	mul_add_c [r1],[r2+12],r3,r4,r5
	move.d r3,[r0+]
	clear.d r3

	mul_add_c [r1+12],[r2+4],r4,r5,r3
	mul_add_c [r1+8],[r2+8],r4,r5,r3
	mul_add_c [r1+4],[r2+12],r4,r5,r3
	move.d r4,[r0+]
	clear.d r4

	mul_add_c [r1+8],[r2+12],r5,r3,r4
	mul_add_c [r1+12],[r2+8],r5,r3,r4
	move.d r5,[r0+]
	clear.d r5

	mul_add_c [r1+12],[r2+12],r3,r4,r5
	move.d r3,[r0+]
	move.d r4,[r0+]

	movem [sp+],r5
	jump [sp+]
	
	.globl _bn_mul_comba8
_bn_mul_comba8:
	;; calculate <r> = <a>*<b>, r is 16 long, a,b are 4 long
	;; register usage:
	;;   r10 - BN_ULONG *r, the result array
	;;   r11 - BN_ULONG *a, source array
	;;   r12 - BN_ULONG *b, source array
	;;   r0 work copy of r10 (<r>)
	;;   r1 work copy of r11 (<a>)
	;;   r2 work copy of r12 (<b>)
	;;   r3 thru r5 carry registers
	;; 
	;;   r6 thru r8 not used
	;; 
	;; r10 thru r13 used for subroutine calling
	;; 

	;; put r0 to r5 on stackframe (6*4==24 octets)
	push srp
	subq 24,sp
	movem r5,[sp]
	;; move indata to r0-r3 since we need r10-r13 for subroutine calls
	move.d r10, r0
	move.d r11, r1
	move.d r12, r2

	clear.d r3
	clear.d r4
	clear.d r5

	mul_add_c [r1],[r2],r3,r4,r5
	move.d r3, [r0+]
	clear.d r3

	mul_add_c [r1],[r2+4],r4,r5,r3
	mul_add_c [r1+4],[r2],r4,r5,r3
	move.d r4, [r0+]
	clear.d r4

	mul_add_c [r1+8],[r2],r5,r3,r4
	mul_add_c [r1+4],[r2+4],r5,r3,r4
	mul_add_c [r1],[r2+8],r5,r3,r4
	move.d r5, [r0+]
	clear.d r5

	mul_add_c [r1],[r2+12],r3,r4,r5
	mul_add_c [r1+4],[r2+8],r3,r4,r5
	mul_add_c [r1+8],[r2+4],r3,r4,r5
	mul_add_c [r1+12],[r2],r3,r4,r5
	move.d r3, [r0+]
	clear.d r3
	
	mul_add_c [r1+16],[r2],r4,r5,r3
	mul_add_c [r1+12],[r2+4],r4,r5,r3
	mul_add_c [r1+8],[r2+8],r4,r5,r3
	mul_add_c [r1+4],[r2+12],r4,r5,r3
	mul_add_c [r1],[r2+16],r4,r5,r3
	move.d r4, [r0+]
	clear.d r4

	mul_add_c [r1],[r2+20],r5,r3,r4
	mul_add_c [r1+4],[r2+16],r5,r3,r4
	mul_add_c [r1+8],[r2+12],r5,r3,r4
	mul_add_c [r1+12],[r2+8],r5,r3,r4
	mul_add_c [r1+16],[r2+4],r5,r3,r4
	mul_add_c [r1+20],[r2],r5,r3,r4
	move.d r5, [r0+]
	clear.d r5

	mul_add_c [r1+24],[r2],r3,r4,r5
	mul_add_c [r1+20],[r2+4],r3,r4,r5
	mul_add_c [r1+16],[r2+8],r3,r4,r5
	mul_add_c [r1+12],[r2+12],r3,r4,r5
	mul_add_c [r1+8],[r2+16],r3,r4,r5
	mul_add_c [r1+4],[r2+20],r3,r4,r5
	mul_add_c [r1],[r2+24],r3,r4,r5
	move.d r3, [r0+]
	clear.d r3

	mul_add_c [r1],[r2+28],r4,r5,r3
	mul_add_c [r1+4],[r2+24],r4,r5,r3
	mul_add_c [r1+8],[r2+20],r4,r5,r3
	mul_add_c [r1+12],[r2+16],r4,r5,r3
	mul_add_c [r1+16],[r2+12],r4,r5,r3
	mul_add_c [r1+20],[r2+8],r4,r5,r3
	mul_add_c [r1+24],[r2+4],r4,r5,r3
	mul_add_c [r1+28],[r2],r4,r5,r3
	move.d r4, [r0+]
	clear.d r4

	mul_add_c [r1+28],[r2+4],r5,r3,r4
	mul_add_c [r1+24],[r2+8],r5,r3,r4
	mul_add_c [r1+20],[r2+12],r5,r3,r4
	mul_add_c [r1+16],[r2+16],r5,r3,r4
	mul_add_c [r1+12],[r2+20],r5,r3,r4
	mul_add_c [r1+8],[r2+24],r5,r3,r4
	mul_add_c [r1+4],[r2+28],r5,r3,r4
	move.d r5, [r0+]
	clear.d r5

	mul_add_c [r1+8],[r2+28],r3,r4,r5
	mul_add_c [r1+12],[r2+24],r3,r4,r5
	mul_add_c [r1+16],[r2+20],r3,r4,r5
	mul_add_c [r1+20],[r2+16],r3,r4,r5
	mul_add_c [r1+24],[r2+12],r3,r4,r5
	mul_add_c [r1+28],[r2+8],r3,r4,r5
	move.d r3, [r0+]
	clear.d r3

	mul_add_c [r1+28],[r2+12],r4,r5,r3
	mul_add_c [r1+24],[r2+16],r4,r5,r3
	mul_add_c [r1+20],[r2+20],r4,r5,r3
	mul_add_c [r1+16],[r2+24],r4,r5,r3
	mul_add_c [r1+12],[r2+28],r4,r5,r3
	move.d r4, [r0+]
	clear.d r4

	mul_add_c [r1+16],[r2+28],r5,r3,r4
	mul_add_c [r1+20],[r2+24],r5,r3,r4
	mul_add_c [r1+24],[r2+20],r5,r3,r4
	mul_add_c [r1+28],[r2+16],r5,r3,r4
	move.d r5, [r0+]
	clear.d r5

	mul_add_c [r1+28],[r2+20],r3,r4,r5
	mul_add_c [r1+24],[r2+24],r3,r4,r5
	mul_add_c [r1+20],[r2+28],r3,r4,r5
	move.d r3, [r0+]
	clear.d r3
	
	mul_add_c [r1+24],[r2+28],r4,r5,r3
	mul_add_c [r1+28],[r2+24],r4,r5,r3
	move.d r4, [r0+]
	clear.d r4

	mul_add_c [r1+28],[r2+28],r5,r3,r4
	move.d r5, [r0+]
	move.d r3, [r0+]

	movem [sp+],r5
	jump [sp+]

	.align 1
	.globl _bn_sqr_comba4
_bn_sqr_comba4:
	;; calculate <r> = <a>**2, r is 8 long, a is 4 long
	;; register usage:
	;;   r10 - BN_ULONG *r, the result array
	;;   r11 - BN_ULONG *a, the source array
	;;   r0 work copy of r10 (<r>)
	;;   r1 work copy of r11 (<a>)
	;;   r2 thru r4 carry registers
	;; 
	;;   r5 thru r8 not used
	;; 
	;; r10 thru r13 used for subroutine calling
	;; 

	;; put r0 to r4 on stackframe (5*4==20 octets)
	push srp
	subq 20,sp
	movem r4,[sp]
	;; move indata to r0-r2 since we need r10-r13 for subroutine calls
	move.d r10, r0
	move.d r11, r1

	.macro sqr_add_c p0,c0,c1,c2
	move.d \p0,r10
	move.d r10,r11
	jsr Lcris_bn_asm_mul_32_to_64

	add.d r10,\c0		  ;
	ax
	addq 0,r11		  ; add in carry, cannot overflow due to
	;;                          size of mul result
	add.d r11,\c1		  ;
	ax
	addq 0,\c2
	.endm

	.macro sqr_add_c2 p0,p1,c0,c1,c2	
	move.d \p0,r10
	move.d \p1,r11
	jsr Lcris_bn_asm_mul_32_to_64

	move.d r10,r12
	move.d r11,r13
	add.d r12,r10
	ax
	add.d r13,r11
	ax
	addq 0,\c2
	add.d r10,\c0
	ax
	addq 0,r11
	ax
	addq 0,\c2
	add.d r11,\c1
	ax
	addq 0,\c2
	.endm

	clear.d r2
	clear.d r3
	clear.d r4

	sqr_add_c [r1],r2,r3,r4
	move.d r2, [r0+]
	clear.d r2

	sqr_add_c2 [r1+4],[r1],r3,r4,r2
	move.d r3, [r0+]
	clear.d r3

	sqr_add_c [r1+4],r4,r2,r3
	sqr_add_c2 [r1+8],[r1],r4,r2,r3
	move.d r4,[r0+]
	clear.d r4

	sqr_add_c2 [r1+12],[r1],r2,r3,r4
	sqr_add_c2 [r1+8],[r1+4],r2,r3,r4
	move.d r2,[r0+]
	clear.d r2

	sqr_add_c [r1+8],r3,r4,r2
	sqr_add_c2 [r1+12],[r1+4],r3,r4,r2
	move.d r3,[r0+]
	clear.d r3

	sqr_add_c2 [r1+12],[r1+8],r4,r2,r3
	move.d r4,[r0+]
	clear.d r4

	sqr_add_c [r1+12],r2,r3,r4
	move.d r2,[r0+]
	move.d r3,[r0+]	
	
	movem [sp+],r4
	jump [sp+]


	.align 1
	.globl _bn_sqr_comba8
_bn_sqr_comba8:
	;; calculate <r> = <a>**2, r is 16 long, a is 4 long
	;; register usage:
	;;   r10 - BN_ULONG *r, the result array
	;;   r11 - BN_ULONG *a, the source array
	;;   r0 work copy of r10 (<r>)
	;;   r1 work copy of r11 (<a>)
	;;   r2 thru r4 carry registers
	;; 
	;;   r5 thru r9 not used
	;; 
	;; r10 thru r13 used for subroutine calling
	;; 

	;; put r0 to r4 on stackframe (5*4==20 octets)
	push srp
	subq 20,sp
	movem r4,[sp]
	;; move indata to r0-r2 since we need r10-r13 for subroutine calls
	move.d r10, r0
	move.d r11, r1

	clear.d r2
	clear.d r3
	clear.d r4

	sqr_add_c [r1],r2,r3,r4
	move.d r2,[r0+]
	clear.d r2
	
	sqr_add_c2 [r1+4],[r1],r3,r4,r2
	move.d r3,[r0+]
	clear.d r3

	sqr_add_c [r1+4],r4,r2,r3
	sqr_add_c2 [r1+8],[r1],r4,r2,r3
	move.d r4,[r0+]
	clear.d r4

	sqr_add_c2 [r1+12],[r1],r2,r3,r4
	sqr_add_c2 [r1+8],[r1+4],r2,r3,r4
	move.d r2,[r0+]
	clear.d r2

	sqr_add_c [r1+8],r3,r4,r2
	sqr_add_c2 [r1+12],[r1+4],r3,r4,r2
	sqr_add_c2 [r1+16],[r1],r3,r4,r2
	move.d r3,[r0+]
	clear.d r3

	sqr_add_c2 [r1+20],[r1],r4,r2,r3
	sqr_add_c2 [r1+16],[r1+4],r4,r2,r3
	sqr_add_c2 [r1+12],[r1+8],r4,r2,r3
	move.d r4,[r0+]
	clear.d r4

	sqr_add_c [r1+12],r2,r3,r4
	sqr_add_c2 [r1+16],[r1+8],r2,r3,r4
	sqr_add_c2 [r1+20],[r1+4],r2,r3,r4
	sqr_add_c2 [r1+24],[r1],r2,r3,r4
	move.d r2,[r0+]
	clear.d r2

	sqr_add_c2 [r1+28],[r1],r3,r4,r2
	sqr_add_c2 [r1+24],[r1+4],r3,r4,r2
	sqr_add_c2 [r1+20],[r1+8],r3,r4,r2
	sqr_add_c2 [r1+16],[r1+12],r3,r4,r2
	move.d r3,[r0+]
	clear.d r3

	sqr_add_c [r1+16],r4,r2,r3
	sqr_add_c2 [r1+20],[r1+12],r4,r2,r3
	sqr_add_c2 [r1+24],[r1+8],r4,r2,r3
	sqr_add_c2 [r1+28],[r1+4],r4,r2,r3
	move.d r4,[r0+]
	clear.d r4

	sqr_add_c2 [r1+28],[r1+8],r2,r3,r4
	sqr_add_c2 [r1+24],[r1+12],r2,r3,r4
	sqr_add_c2 [r1+20],[r1+16],r2,r3,r4
	move.d r2,[r0+]
	clear.d r2

	sqr_add_c [r1+20],r3,r4,r2
	sqr_add_c2 [r1+24],[r1+16],r3,r4,r2
	sqr_add_c2 [r1+28],[r1+12],r3,r4,r2
	move.d r3,[r0+]
	clear.d r3

	sqr_add_c2 [r1+28],[r1+16],r4,r2,r3
	sqr_add_c2 [r1+24],[r1+20],r4,r2,r3
	move.d r4,[r0+]
	clear.d r4

	sqr_add_c [r1+24],r2,r3,r4
	sqr_add_c2 [r1+28],[r1+20],r2,r3,r4
	move.d r2,[r0+]
	clear.d r2

	sqr_add_c2 [r1+28],[r1+24],r3,r4,r2
	move.d r3,[r0+]
	clear.d r3

	sqr_add_c [r1+28],r4,r2,r3
	move.d r4,[r0+]
	move.d r2,[r0+]
	
	;; end
	movem [sp+],r4
	jump [sp+]

	;; snagged from gcc-cris output
	.align 1
	.global _bn_div_words
_bn_div_words:
	Push srp
	subq 24,sp
	movem r5,[sp]
	move.d r10,r0
	clear.d r1
	move.d r0,r5
	clear.d r4
	move.d r11,r9
	clear.d r10
	move.d r9,r0
	move.d r5,r1
	move.d r12,r2
	clear.d r3
	move.d r0,r10
	move.d r5,r11
	move.d r2,r12
	clear.d r13
	Jsr ___udivdi3
	move.d r10,r12
	move.d r11,r13
	move.d r12,r10
	movem [sp+],r5
	jump [sp+]

	
	;; multiplication subroutine, multiply two 32 bit unsigneds into
	;; a 64 bit unsigned result
	;; in:	r10 (a),r11 (b)
	;; out:	r10 (low dword), r11 (high dword)
	.align 1
	;; 	.global _cris_bn_asm_mul_32_to_64
Lcris_bn_asm_mul_32_to_64:
	;; r0  16 bit partial result 1
	;; r1  16 bit partial result 2
	;; r13 tmp store of low temporary result
	;; r12 tmp store of mid temporary result
	;; r9 tmp store of high temporary result

	subq 8,sp
	movem r1,[sp]

	;; multiply a_low * b_low, put result in r13
	clear.d r0
	move.w r10,r0
	move.d r11,r13
	lslq 16,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13
	mstep r0,r13

	;; multiply a_low * b_high, put result in r12
	move.d r11,r12
	clear.w r12
	test.d r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12
	mstep r0,r12

	;; multiply a_high * b_low
	move.d r10,r1
	clear.d r0
	move.w r11,r0
	clear.w r1
	test.d r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	clear.d r9
	add.d r1,r12		; add to first mid temporary result
	ax
	addq 0,r9
	lslq 16,r9

	;; multiply a_high * b_high, put result in r9
	move.d r10,r0
	move.d r11,r1
	lsrq 16,r0		; zero extend, high word is cleared
	clear.w r1
	test.d r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	mstep r0,r1
	add.d r1,r9

	;; calc return dwords
	move.d r9,r11		; clear hi res dword
	move.d r12,r10		; move low word of mid result to high
	lslq 16,r10		;    word of low return dword
	lsrq 16,r12
	add.d r13,r10		; add low result to low return val
	ax
	addq 0,r11		; add carry to high return val
	add.d r12,r11		; add high word of mid result

	movem [sp+],r1
	Ret
	nop
